@@ -1,9 +1,9 @@
 #define SPRAY
 #ifndef lint
 static char *rcsid =
-   "@(#)$Header: trACESroute.c,v 6.0 GOLD 1999/11/15 gavron(gavron@aces.com)";
+   "@(#)$Header: trACESroute.c,v 6.1.1 GOLD 2000/04/29 gavron(gavron@aces.com)";
 #endif
-#define TR_VERSION "6.0 GOLD"
+#define TR_VERSION "6.1.1 GOLD"
 
 /*
  * traceroute host  - trace the route ip packets follow going to "host".
@@ -449,6 +449,11 @@ static char *rcsid =
  *			to avoid 'confusion' with vj/lbl traceroute 2.9.x.
  *			Of course this makes us much more advanced than
  *			Netscape 4.8, I.E. 5.0, or The New AOL 5.0. ;-)
+ * 01/18/2000   V6.1    Dan Cohn <dan@internap.com> provided mods to make
+ *                      spray mode as functional as normal mode, both in
+ *                      terms of missing packets and in multiple probes.
+ * 04/29/2000	V6.1.1	As per Thomas Erskine, add info for -A where the
+ *			route is not in the RR server queried.
  */
 
 #include <stdio.h>
@@ -585,7 +590,7 @@ int gettimeofday(struct timeval *, void *);
 #include <netinet/ip_var.h>
 #else /* __linux__ */
 #include <sys/time.h>
-#include <netinet/if_tr.h>
+/* IRD #include <netinet/if_tr.h> */
 #endif /* __linux__ */
 #include <netinet/ip_icmp.h>
 #include <netinet/udp.h>
@@ -683,7 +688,7 @@ struct {
         struct  timeval rtn;    /* time packet arrived */
         struct  sockaddr_in from; /* whom from */
 } spray[SPRAYMAX];
-int spray_rtn[SPRAYMAX];	/* See which TTLs have responded */
+int *spray_rtn[SPRAYMAX];	/* See which TTLs have responded */
 int spray_target;		/* See which TTL the host responds on */
 int spray_max;			/* See which is the highest TTL we've seen */
 int spray_min;			/* See smallest host-returned TTL */
@@ -748,6 +753,7 @@ int dns_owner_lookup=0;         /* Look up owner email in DNS */
 int as_lookup=0;                /* Look up AS path in routing registries */
 int got_there;
 int unreachable;
+int response_mask = 0;
 int mtu, new_mtu = 0;
 
 char usage[] = "%s: TrACESroute Usage: traceroute [-adnruvAMOQ] [-w wait] [-S start_ttl] [-m max_ttl] [-p port#] [-q nqueries] [-g gateway] [-t tos] [-s src_addr] [-g router] host [data size]\n\
@@ -1298,8 +1304,9 @@ main(argc, argv)
 
    /* For all TTL do */
    for (ttl = min_ttl; ttl <= max_ttl; ++ttl) {
-      spray_rtn[ttl]=0;
+      spray_rtn[ttl] = (int *)malloc(sizeof(int)*nprobes + 1);
       for (probe = 0; probe < nprobes; ++probe) {
+	     spray_rtn[ttl][probe]=0;
          send_probe(++seq, ttl);
       }
    }
@@ -1311,7 +1318,7 @@ main(argc, argv)
       deadline.tv_sec++;
    }
    /* Go get responses until either we get them all, or timeout */
-   while (  (spray_min > spray_total)  &&
+   while (  ((((int)pow(2,spray_min)-1)&response_mask) != ((int)pow(2,spray_min)-2)) &&
             (cc = wait_for_reply(s,&from,&deadline))  ) {
       (void) packet_ok(packet, cc, &from, seq);
    }
@@ -1319,36 +1326,52 @@ main(argc, argv)
    last_i = 1;
    last_ttl = 0;
    bzero(&addr_last, sizeof(addr_last));
-   for (i = 1; i <= spray_min; i++) {
+   for (i = 1; i <= max_ttl; i++) {
       /* First see if it's valid, and if so play with its time */
-      idx = spray_rtn[i];
-      if ((idx > 0) && (idx < SPRAYMAX)) {
-	 spr_ttl = spray[idx].ttl;
-	 if (spr_ttl != i) {
-	    Fprintf(stderr,"Check failure spray(rtn[i]) !=i\n");
-	    exit(0);
-	 }
-	 ttl_diff = (spr_ttl - last_ttl);
-	 if (ttl_diff > 1) {
-            for (last_i = last_ttl+1; last_i < spr_ttl; last_i++) {
-               print_ttl(last_i);
-	       for (probe = 1; probe <= nprobes ; probe++) { 
-                  Fprintf(stdout," *");
-               } /* end for probes */
-               Fprintf(stdout,terminator);
-            } /* end for last i */
-	 } /* endf if spr_ttl ... */
-         if (spr_ttl > last_ttl) {
-            print_ttl(spr_ttl);
-            last_ttl = spr_ttl;
-         }
+      idx = spray_rtn[i][0];
+      if ((idx > 0) && (idx < SPRAYMAX) && (spray[idx].from.sin_addr.s_addr != 0)) { 
+	      spr_ttl = spray[idx].ttl;
+	      if (spr_ttl != i) {
+	          Fprintf(stderr,"Check failure spray(rtn[i]) !=i\n");
+	          exit(0);
+	      }
 
-         print_from(&spray[idx].from);
-         tvsub(&spray[spray_rtn[i]].rtn,&spray[spray_rtn[i]].out);
-         ddt=spray[spray_rtn[i]].rtn.tv_sec*1000.0+
-            ((float)spray[spray_rtn[i]].rtn.tv_usec)/1000.0;
-	 print_time(&ddt);
-         Fprintf(stdout,terminator);
+          /* do not display duplicate entries (responses beyond terminus from same host) */
+	      if (0 != memcmp( &spray[idx].from.sin_addr.s_addr,
+                          &addr_last.sin_addr.s_addr,
+		                  sizeof(addr_last.sin_addr.s_addr) )) {
+
+        	 ttl_diff = (spr_ttl - last_ttl);
+        	 if (ttl_diff > 1) {
+                   for (last_i = last_ttl+1; last_i < spr_ttl; last_i++) {
+                       print_ttl(last_i);
+        	       for (probe = 1; probe <= nprobes ; probe++) { 
+                          Fprintf(stdout," *");
+                       } /* end for probes */
+                       Fprintf(stdout,terminator);
+                   } /* end for last i */
+        	 } /* endf if spr_ttl ... */
+        
+             if (spr_ttl > last_ttl) {
+                print_ttl(spr_ttl);
+                last_ttl = spr_ttl;
+             }
+
+             print_from(&spray[idx].from);
+
+             for (probe = 0; probe < nprobes; ++probe) {
+    		     if (spray_rtn[i][probe] != 0) {
+                     tvsub(&spray[spray_rtn[i][probe]].rtn,&spray[spray_rtn[i][probe]].out);
+                     ddt=spray[spray_rtn[i][probe]].rtn.tv_sec*1000.0+
+                        ((float)spray[spray_rtn[i][probe]].rtn.tv_usec)/1000.0;
+                     print_time(&ddt);
+    			 } else {
+    			     Fprintf(stdout," *");
+    			 }
+             }
+
+             Fprintf(stdout,terminator);
+		 } /* no duplicate entries */
       } /* end if nonzero type */
    } /* end for */
 } /* end spray mode */
@@ -1588,6 +1611,7 @@ packet_ok(buf, cc, from, seq)
 	int tmp,tmp2;
 	int spr_seq;
 	int spr_ttl;
+	int probecnt;
 	
 #ifndef ARCHAIC
 	struct ip *ip;
@@ -1630,24 +1654,32 @@ packet_ok(buf, cc, from, seq)
 		   spr_seq = ntohs(up->uh_dport)-port;
 	   	   if ( (spr_seq >=0) && (spr_seq < max_ttl*nprobes+1) ) {
                       spr_ttl = spray[spr_seq].ttl;
+					  response_mask |= (1 << spr_ttl);
 /*
  * Now we increment the response count for this ttl, and then if it's the 
  * first, increment the total of ttl's seen 
  */
-		      if (spray_rtn[spr_ttl] == 0) {
-                         spray_total++;
-		      }
-		      spray_rtn[spr_ttl] = spr_seq;
+		      if (spray_rtn[spr_ttl][0] == 0) {
+                        spray_total++;
+		      } 
+
+              for (probecnt = 0; probecnt < nprobes; ++probecnt) {
+		          if (spray_rtn[spr_ttl][probecnt] == 0) {
+				      spray_rtn[spr_ttl][probecnt] = spr_seq;
+					  break;
+				  }
+			  }
 /*
  * We want to do some heuristics on the smallest TTL received from the target
  * host, but we need the type for that...
  */
-                      spray[spr_seq].type = type;
-                      if (type == ICMP_UNREACH_PORT) {
-		         if (spr_ttl < spray_min) 
-			    spray_min = spr_ttl;
-			    spray_target = spr_ttl;
-		      }
+                spray[spr_seq].type = type;
+                if (type == ICMP_UNREACH_PORT) {
+		            if (spr_ttl < spray_min)  {
+			            spray_min = spr_ttl;
+				    }
+			        spray_target = spr_ttl;
+		        }
 /*
  * We also want the largest TTL we've seen...
  */
@@ -1934,6 +1966,7 @@ struct in_addr in;
  * an example for the truly clueless...
  *
  *	20-May-1995	Ehud Gavron	gavron@aces.com
+ *	28-Apr-2000			Return error if no string
  */
 
 
@@ -2128,6 +2161,14 @@ char *net;
 
 	reply[count] = '\0';	/* Terminate it - thanks Joey! */
 
+  	/*
+   	 * sometimes there's no answer
+	 */
+	if (strncmp(reply, "%%  No entries found for the selected source(s).",
+	    strlen("%%  No entries found for the selected source(s).")) ==0) {
+	   return "NONE";
+	}
+
 	/* 
 	 * So now we have a large string, somewhere in which we can
 	 * find  origin:*AS%%%%%%<lf>.  We parse this into AS%%%%%.
